@@ -23,7 +23,7 @@ impl Column for PgColumn {
     }
 
     fn name(&self) -> &str {
-        &*self.name
+        &self.name
     }
 
     fn type_info(&self) -> &PgTypeInfo {
